@@ -29,6 +29,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <stdarg.h>
 
 #include "ratpoison.h"
 
@@ -59,11 +60,12 @@ show_bar (screen_info *s)
   if (!s->bar_is_raised)
     {
       s->bar_is_raised = BAR_IS_WINDOW_LIST;
-      XMapWindow (dpy, s->bar_window);
+      XMapRaised (dpy, s->bar_window);
       update_window_names (s);
   
       /* Set an alarm to auto-hide the bar BAR_TIMEOUT seconds later */
       alarm (BAR_TIMEOUT);
+      alarm_signalled = 0;
       return 1;
     }
 
@@ -91,54 +93,46 @@ void
 update_window_names (screen_info *s)
 {
   static struct sbuf *bar_buffer = NULL;
-  rp_window *w;
-  char dbuf[10];
 
   int mark_start = 0;
   int mark_end = 0;
 
-  if (!s->bar_is_raised) return;
+  if (s->bar_is_raised != BAR_IS_WINDOW_LIST) return;
 
   if (bar_buffer == NULL)
     bar_buffer = sbuf_new (0);
 
-  sbuf_clear (bar_buffer);
-
-  for (w = rp_mapped_window_sentinel->next; 
-       w != rp_mapped_window_sentinel; 
-       w = w->next)
-    {
-      PRINT_DEBUG ("%d-%s\n", w->number, w->name);
-
-      if (w == rp_current_window)
-	mark_start = strlen (sbuf_get (bar_buffer));
-
-      sbuf_concat (bar_buffer, " ");
-
-      sprintf (dbuf, "%d", w->number);
-      sbuf_concat (bar_buffer, dbuf);
-
-      if (w == rp_current_window)
-	sbuf_concat (bar_buffer, "*");
-      else
-	sbuf_concat (bar_buffer, "-");
-
-      sbuf_concat (bar_buffer, w->name);
-
-      sbuf_concat (bar_buffer, " ");
-
-      if (w == rp_current_window)
-	mark_end = strlen (sbuf_get (bar_buffer));
-    }
-
-  if (!strcmp (sbuf_get (bar_buffer), ""))
-    {
-      sbuf_copy (bar_buffer, MESSAGE_NO_MANAGED_WINDOWS);
-    }
+  get_window_list (NULL, bar_buffer, &mark_start, &mark_end);
 
   marked_message (sbuf_get (bar_buffer), mark_start, mark_end);
 }
 
+void
+marked_message_printf (int mark_start, int mark_end, char *fmt, ...)
+{
+  int size, nchars;
+  char *buffer;
+  va_list ap;
+
+  va_start (ap, fmt);
+
+  /* A resonable starting value. */
+  size = strlen (fmt) + 1;
+  buffer = (char *)xmalloc (size);
+
+  nchars = vsnprintf (buffer, size, fmt, ap);
+  if (nchars >= size)
+    {
+      buffer = (char *)xrealloc (buffer, nchars + 1);
+      vsnprintf (buffer, nchars + 1, fmt, ap);
+    }
+
+  va_end (ap);
+
+  marked_message (buffer, mark_start, mark_end);
+  free (buffer);
+}
+
 void
 marked_message (char *msg, int mark_start, int mark_end)
 {
@@ -156,11 +150,12 @@ marked_message (char *msg, int mark_start, int mark_end)
   if (!s->bar_is_raised)
     {
       s->bar_is_raised = BAR_IS_MESSAGE;
-      XMapWindow (dpy, s->bar_window);
+      XMapRaised (dpy, s->bar_window);
     }
 
   /* Reset the alarm to auto-hide the bar in BAR_TIMEOUT seconds. */
   alarm (BAR_TIMEOUT);
+  alarm_signalled = 0;
 
   XMoveResizeWindow (dpy, s->bar_window, 
 		     bar_x (s, width), bar_y (s),
@@ -175,19 +170,36 @@ marked_message (char *msg, int mark_start, int mark_end)
 	       BAR_Y_PADDING + s->font->max_bounds.ascent,
 	       msg, strlen (msg));
 
+
+  /* Crop to boundary conditions. */
+  if (mark_start < 0)
+    mark_start = 0;
+
+  if (mark_end < 0)
+    mark_end = 0;
+
+  if (mark_start > strlen (msg))
+    mark_start = strlen (msg);
+
+  if (mark_end > strlen (msg))
+    mark_end = strlen (msg);
+
+  if (mark_start > mark_end)
+    {
+      int tmp;
+      tmp = mark_start;
+      mark_start = mark_end;
+      mark_end = tmp;
+    }
+
   /* xor the string representing the current window */
   if (mark_start != mark_end)
     {
       int start;
       int end;
 
-      assert (mark_start <= mark_end); /* FIXME: remove these assertions,
-				      make it work in all cases */
-      assert (mark_start <= strlen(msg) + 1);
-      assert (mark_end <= strlen(msg) + 1);
-      
       start = XTextWidth (s->font, msg, mark_start) + BAR_X_PADDING;
-      end = XTextWidth (s->font, msg + mark_start, mark_end - mark_start);
+      end = XTextWidth (s->font, msg + mark_start, mark_end - mark_start) + BAR_X_PADDING;
 
       PRINT_DEBUG ("%d %d strlen(%d)==> %d %d\n", mark_start, mark_end, strlen(msg), start, end);
 
