@@ -28,13 +28,21 @@
 
 #define FONT_HEIGHT(f) ((f)->max_bounds.ascent + (f)->max_bounds.descent)
 
-#define STATE_UNMAPPED 0
-#define STATE_MAPPED   1
-
+#define WIN_EVENTS (StructureNotifyMask | PropertyChangeMask | ColormapChangeMask | FocusChangeMask)
 
 typedef struct rp_window rp_window;
 typedef struct screen_info screen_info;
 typedef struct rp_action rp_action;
+typedef struct rp_window_frame rp_window_frame;
+
+struct rp_window_frame
+{
+  int x, y, width, height;
+  rp_window *win;
+
+  rp_window_frame *prev, *next;
+};
+
 
 struct rp_window
 {
@@ -45,7 +53,7 @@ struct rp_window
   int state;
   int last_access;		
   int named;
-  
+
   /* Dimensions */
   int x, y, width, height, border;
 
@@ -62,6 +70,11 @@ struct rp_window
   /* Saved mouse position */
   int mouse_x, mouse_y;
 
+  /* A window can be visible inside a frame but not the frame's
+     current window. This keeps track of what frame the window was
+     mapped into. */
+  rp_window_frame *frame;
+
   rp_window *next, *prev;  
 };
 
@@ -70,10 +83,11 @@ struct screen_info
   GC normal_gc;
   XFontStruct *font;		/* The font we want to use. */
   XWindowAttributes root_attr;
-  Window root, bar_window, key_window, input_window;
+  Window root, bar_window, key_window, input_window, frame_window, help_window;
   int bar_is_raised;
   int screen_num;		/* Our screen number as dictated my X */
   Colormap def_cmap;
+  Cursor rat;
 };
 
 struct rp_action
@@ -111,18 +125,25 @@ extern rp_window *rp_mapped_window_sentinel;
    assigned to them and are not visible/active. */
 extern rp_window *rp_unmapped_window_sentinel;
 
-/* Pointer to the currently focused window. Points to an element in
-   the mapped window list. NULL if there are no mapped windows. */
-extern rp_window *rp_current_window;
+/* A list of frames that may or may not contain windows. There should
+   always be one in the list. */
+extern rp_window_frame *rp_window_frame_sentinel;
+
+/* Pointer to the currently focused frame. */
+extern rp_window_frame *rp_current_frame;
 
 extern screen_info *screens;
 extern int num_screens;
 
-extern XEvent *rp_current_event;
+extern XEvent rp_current_event;
 
 extern Display *dpy;
 extern Atom rp_restart;
 extern Atom rp_kill;
+extern Atom rp_command;
+extern Atom rp_command_request;
+extern Atom rp_command_result;
+
 
 extern Atom wm_state;
 extern Atom wm_change_state;
@@ -151,9 +172,26 @@ struct modifier_info
   unsigned int alt_mod_mask;
   unsigned int super_mod_mask;
   unsigned int hyper_mod_mask;
+
+  /* Keep track of these because they mess up the grab and should be
+     ignored. */  
+  unsigned int num_lock_mask;
+  unsigned int scroll_lock_mask; 
 };
 
 /* Keeps track of which mod mask each modifier is under. */
 extern struct modifier_info rp_modifier_info;
 
+/* nonzero if an alarm signal was raised. This means ratpoison should
+   hide its popup windows. */
+extern int alarm_signalled;
+extern int kill_signalled;
+extern int hup_signalled;
+
+/* rudeness levels */
+extern int rp_honour_transient_raise;
+extern int rp_honour_normal_raise;
+extern int rp_honour_transient_map;
+extern int rp_honour_normal_map;
+
 #endif /* _RATPOISON_DATA_H */
