@@ -23,9 +23,4 @@
 #define _RATPOISON_EVENTS_H 1
 
 void handle_events ();
-void delegate_event (XEvent *ev);
-void key_press (XEvent *ev);
-void map_request (XEvent *ev);
-void unmap_notify (XEvent *ev);
-void keymap_state (XKeymapEvent *ev);
 #endif /* _RATPOISON_EVENTS_H */
