@@ -30,18 +30,19 @@
 
 /* Figure out what keysyms are attached to what modifiers */
 void
-init_modifier_map ()
+update_modifier_map ()
 {
   unsigned int modmasks[] = 
     { Mod1Mask, Mod2Mask, Mod3Mask, Mod4Mask, Mod5Mask };
   int row, col;	/* The row and column in the modifier table.  */
   XModifierKeymap *mods;
 
-/*   rp_modifier_info.mode_switch_mask = 0; */
   rp_modifier_info.meta_mod_mask = 0;
   rp_modifier_info.alt_mod_mask = 0;
   rp_modifier_info.super_mod_mask = 0;
   rp_modifier_info.hyper_mod_mask = 0;
+  rp_modifier_info.num_lock_mask = 0;
+  rp_modifier_info.scroll_lock_mask = 0;
 
   mods = XGetModifierMapping (dpy);
 
@@ -54,40 +55,45 @@ init_modifier_map ()
 	
 	switch (XKeycodeToKeysym(dpy, code, 0))
 	  {
-	    case XK_Meta_L:
-	    case XK_Meta_R:
-	      rp_modifier_info.meta_mod_mask |= modmasks[row - 3];
-	      PRINT_DEBUG ("Found Meta on %d\n",
-			   rp_modifier_info.meta_mod_mask);
-	      break;
+	  case XK_Meta_L:
+	  case XK_Meta_R:
+	    rp_modifier_info.meta_mod_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found Meta on %d\n",
+			 rp_modifier_info.meta_mod_mask);
+	    break;
 
-	    case XK_Alt_L:
-	    case XK_Alt_R:
-	      rp_modifier_info.alt_mod_mask |= modmasks[row - 3];
-	      PRINT_DEBUG ("Found Alt on %d\n",
-			   rp_modifier_info.alt_mod_mask);
-	      break;
+	  case XK_Alt_L:
+	  case XK_Alt_R:
+	    rp_modifier_info.alt_mod_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found Alt on %d\n",
+			 rp_modifier_info.alt_mod_mask);
+	    break;
 
-	    case XK_Super_L:
-	    case XK_Super_R:
-	      rp_modifier_info.super_mod_mask |= modmasks[row - 3];
-	      PRINT_DEBUG ("Found Super on %d\n",
-			   rp_modifier_info.super_mod_mask);
-	      break;
+	  case XK_Super_L:
+	  case XK_Super_R:
+	    rp_modifier_info.super_mod_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found Super on %d\n",
+			 rp_modifier_info.super_mod_mask);
+	    break;
 
-	    case XK_Hyper_L:
-	    case XK_Hyper_R:
-	      rp_modifier_info.hyper_mod_mask |= modmasks[row - 3];
-	      PRINT_DEBUG ("Found Hyper on %d\n",
-			   rp_modifier_info.hyper_mod_mask);
-	      break;
+	  case XK_Hyper_L:
+	  case XK_Hyper_R:
+	    rp_modifier_info.hyper_mod_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found Hyper on %d\n",
+			 rp_modifier_info.hyper_mod_mask);
+	    break;
 
-/* 	  case XK_Mode_switch: */
-/* 	    rp_modifier_info.mode_switch_mask |= modmasks[row - 3]; */
-/* 	      PRINT_DEBUG ("Found Mode_switch on %d\n", */
-/* 			   rp_modifier_info.mode_switch_mask); */
-/* 	    break; */
+	  case XK_Num_Lock:
+	    rp_modifier_info.num_lock_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found NumLock on %d\n", 
+			 rp_modifier_info.num_lock_mask);
+	    break;
 
+	  case XK_Scroll_Lock:
+	    rp_modifier_info.scroll_lock_mask |= modmasks[row - 3];
+	    PRINT_DEBUG ("Found ScrollLock on %d\n", 
+			 rp_modifier_info.scroll_lock_mask);
+	    break;
 	  default:
 	    break;
 	  }
@@ -111,37 +117,56 @@ init_modifier_map ()
   XFreeModifiermap (mods);
 }
 
+/* Grab the key while ignoring annoying modifier keys including
+   caps lock, num lock, and scroll lock. */
+void
+grab_key (int keycode, unsigned int modifiers, Window grab_window)
+{
+  unsigned int mod_list[8];
+  int i;
+
+  /* Create a list of all possible combinations of ignored
+     modifiers. Assumes there are only 3 ignored modifiers. */
+  mod_list[0] = 0;
+  mod_list[1] = LockMask;
+  mod_list[2] = rp_modifier_info.num_lock_mask;
+  mod_list[3] = mod_list[1] | mod_list[2];
+  mod_list[4] = rp_modifier_info.scroll_lock_mask;
+  mod_list[5] = mod_list[1] | mod_list[4];
+  mod_list[6] = mod_list[2] | mod_list[4];
+  mod_list[7] = mod_list[1] | mod_list[2] | mod_list[4];
+
+  /* Grab every combination of ignored modifiers. */
+  for (i=0; i<8; i++)
+    {
+      XGrabKey(dpy, keycode, modifiers | mod_list[i],
+	       grab_window, True, GrabModeAsync, GrabModeAsync);
+    }
+}
+
+
 /* Return the name of the keysym. caller must free returned pointer */
 char *
 keysym_to_string (KeySym keysym, unsigned int modifier)
 {
-  const unsigned int mod_table[] = {ControlMask, 
-				    Mod1Mask, 
-				    Mod2Mask, 
-				    Mod3Mask, 
-				    Mod4Mask, 
-				    Mod5Mask};
-  const unsigned char str_table[] = "CM2345";
+  struct sbuf *name;
+  char *tmp;
 
-  unsigned char *name;
-  int pos, i;
+  name = sbuf_new (0);
 
-  name = xmalloc (15);
+  if (modifier & ControlMask) sbuf_concat (name, "C-");
+  if (modifier & rp_modifier_info.meta_mod_mask) sbuf_concat (name, "M-");
+  if (modifier & rp_modifier_info.alt_mod_mask) sbuf_concat (name, "A-");
+  if (modifier & rp_modifier_info.hyper_mod_mask) sbuf_concat (name, "H-");
+  if (modifier & rp_modifier_info.super_mod_mask) sbuf_concat (name, "S-");
+    
+  sbuf_concat (name, XKeysymToString (keysym));
 
-  for (pos = 0, i = 0; i < 6; i++)
-    {
-      if (modifier & mod_table[i])
-	{
-	  name[pos] = str_table[i];
-	  name[pos+1] = '-';
-	  pos += 2;
-	}
-    }
+  /* Eat the nut and throw away the shells. */
+  tmp = sbuf_get (name);
+  free (name);
 
-  name[pos] = keysym;
-  name[pos+1] = '\0';
-
-  return name;
+  return tmp;
 }
 
 /* Cooks a keycode + modifier into a keysym + modifier. This should be
@@ -152,11 +177,18 @@ keysym_to_string (KeySym keysym, unsigned int modifier)
    interested in the keysym name pass in NULL for keysym_name and 0
    for len. */
 int
-cook_keycode (XKeyEvent *ev, KeySym *keysym, unsigned int *mod, char *keysym_name, int len)
+cook_keycode (XKeyEvent *ev, KeySym *keysym, unsigned int *mod, char *keysym_name, int len, int ignore_bad_mods)
 {
-int nbytes;
+  int nbytes;
+ 
+  if (ignore_bad_mods)
+    {
+      ev->state &= ~(LockMask
+		     | rp_modifier_info.num_lock_mask
+		     | rp_modifier_info.scroll_lock_mask);
+    }
 
- nbytes =  XLookupString (ev, keysym_name, len, keysym, NULL);
+  nbytes =  XLookupString (ev, keysym_name, len, keysym, NULL);
 
   *mod = ev->state;
   *mod &= (rp_modifier_info.meta_mod_mask
@@ -243,7 +275,7 @@ read_key (KeySym *keysym, unsigned int *modifiers, char *keysym_name, int len)
     {
       XMaskEvent (dpy, KeyPressMask, &ev);
       *modifiers = ev.xkey.state;
-      nbytes = cook_keycode (&ev.xkey, keysym, modifiers, keysym_name, len);
+      nbytes = cook_keycode (&ev.xkey, keysym, modifiers, keysym_name, len, 0);
     } while (IsModifierKey (*keysym));
 
   return nbytes;
@@ -332,6 +364,14 @@ get_more_input (char *prompt, char *preinput)
 	  if (cur_len > 0) cur_len--;
 	  update_input_window(s, prompt, str, cur_len);
 	}
+      else if (ch == INPUT_ABORT_KEY && modifier == INPUT_ABORT_MODIFIER)
+	{
+	  /* User aborted. */
+	  free (str);
+	  XSetInputFocus (dpy, fwin, RevertToPointerRoot, CurrentTime);
+	  XUnmapWindow (dpy, s->input_window);
+	  return NULL;
+	}
       else
 	{
 	  if (cur_len + nbytes > allocated_len - 1)
