@@ -23,10 +23,11 @@
 #define _RATPOISON_INPUT_H 1
 
 char *keysym_to_string (KeySym keysym, unsigned int modifier);
-int cook_keycode (XKeyEvent *ev, KeySym *keysym, unsigned int *mod, char *keysym_name, int len);
+int cook_keycode (XKeyEvent *ev, KeySym *keysym, unsigned int *mod, char *keysym_name, int len, int ignore_bad_mods);
 char *get_input (char *prompt);
 char *get_more_input (char *prompt, char *preinput);
 int read_key (KeySym *keysym, unsigned int *modifiers, char *keysym_name, int len);
-void init_modifier_map ();
+void update_modifier_map ();
+void grab_key (int keycode, unsigned int modifiers, Window grab_window);
 
 #endif /* ! _RATPOISON_INPUT_H */
