@@ -28,7 +28,6 @@
 
 rp_window *rp_unmapped_window_sentinel;
 rp_window *rp_mapped_window_sentinel;
-rp_window *rp_current_window;
 
 /* Get the mouse position relative to the root of the specified window */
 static void
@@ -41,6 +40,18 @@ get_mouse_root_position (rp_window *win, int *mouse_x, int *mouse_y)
   XQueryPointer (dpy, win->scr->root, &root_win, &child_win, &win_x, &win_y, mouse_x, mouse_y, &mask);
 }
 
+void
+free_window (rp_window *w)
+{
+  if (w == NULL) return;
+
+  free (w->name);
+  XFree (w->hints);
+  
+  free (w);
+}
+
+
 /* Allocate a new window and add it to the list of managed windows */
 rp_window *
 add_to_window_list (screen_info *s, Window w)
@@ -53,7 +64,7 @@ add_to_window_list (screen_info *s, Window w)
   new_window->scr = s;
   new_window->last_access = 0;
   new_window->prev = NULL;
-  new_window->state = STATE_UNMAPPED;
+  new_window->state = WithdrawnState;
   new_window->number = -1;	
   new_window->named = 0;
   new_window->hints = XAllocSizeHints ();
@@ -63,7 +74,7 @@ add_to_window_list (screen_info *s, Window w)
 
   get_mouse_root_position (new_window, &new_window->mouse_x, &new_window->mouse_y);
 
-  XSelectInput (dpy, new_window->w, PropertyChangeMask | ColormapChangeMask);
+  XSelectInput (dpy, new_window->w, WIN_EVENTS);
 
   new_window->name = xmalloc (strlen ("Unnamed") + 1);
 
@@ -105,12 +116,10 @@ find_window (Window w)
   return win;
 }
 
-
-
 void
 set_current_window (rp_window *win)
 {
-  rp_current_window = win;  
+  set_frames_window (rp_current_frame, win);
 }
 
 void
@@ -124,8 +133,6 @@ init_window_list ()
 
   rp_unmapped_window_sentinel->next = rp_unmapped_window_sentinel;
   rp_unmapped_window_sentinel->prev = rp_unmapped_window_sentinel;
-
-  rp_current_window = NULL;
 }
 
 rp_window *
@@ -178,50 +185,52 @@ find_window_name (char *name)
   return NULL;
 }
 
-/* return the previous window in the list. Assumes window is in the
+/* Return the previous window in the list. Assumes window is in the
    mapped window list. */
 rp_window*
 find_window_prev (rp_window *w)
 {
-  rp_window *prev;
+  rp_window *cur;
 
   if (!w) return NULL;
 
-  prev = w->prev;
-  
-  if (prev == rp_mapped_window_sentinel) 
+  for (cur = w->prev; 
+       cur != w;
+       cur = cur->prev)
     {
-      prev = rp_mapped_window_sentinel->prev;
-      if (prev == w)
+      if (cur == rp_mapped_window_sentinel) continue;
+
+      if (!find_windows_frame (cur))
 	{
-	  return NULL;
+	  return cur;
 	}
     }
 
-  return prev;
+  return NULL;
 }
 
-/* return the next window in the list. Assumes window is in the mapped
+/* Return the next window in the list. Assumes window is in the mapped
    window list. */
 rp_window*
 find_window_next (rp_window *w)
 {
-  rp_window *next;
+  rp_window *cur;
 
   if (!w) return NULL;
 
-  next = w->next;
-  
-  if (next == rp_mapped_window_sentinel) 
+  for (cur = w->next; 
+       cur != w;
+       cur = cur->next)
     {
-      next = rp_mapped_window_sentinel->next;
-      if (next == w)
+      if (cur == rp_mapped_window_sentinel) continue;
+
+      if (!find_windows_frame (cur))
 	{
-	  return NULL;
+	  return cur;
 	}
     }
 
-  return next;
+  return NULL;
 }
 
 rp_window *
@@ -236,7 +245,8 @@ find_window_other ()
        cur = cur->next)
     {
       if (cur->last_access >= last_access 
-	  && cur != rp_current_window)
+	  && cur != current_window()
+	  && !find_windows_frame (cur))
 	{
 	  most_recent = cur;
 	  last_access = cur->last_access;
@@ -351,51 +361,265 @@ save_mouse_position (rp_window *win)
      what we store in mouse_x and mouse_y since they will never be
      used again. */
 
-  ignore_badwindow = 1;
+  ignore_badwindow++;
 
   XQueryPointer (dpy, win->w, &root_win, &child_win, 
 		 &win->mouse_x, &win->mouse_y, &win_x, &win_y, &mask);
 
-  ignore_badwindow = 0;
+  ignore_badwindow--;
+}
+
+/* Takes focus away from last_win and gives focus to win */
+void
+give_window_focus (rp_window *win, rp_window *last_win)
+{
+  /* counter increments every time this function is called. This way
+     we can track which window was last accessed. */
+  static int counter = 1;
+
+  if (win == NULL) return;
+
+  counter++;
+  win->last_access = counter;
+
+  unhide_window (win);
+
+  /* Warp the cursor to the window's saved position. */
+  if (last_win != NULL) save_mouse_position (last_win);
+  XWarpPointer (dpy, None, win->scr->root, 
+		0, 0, 0, 0, win->mouse_x, win->mouse_y);
+  
+  /* Swap colormaps */
+  if (last_win != NULL) XUninstallColormap (dpy, last_win->colormap);
+  XInstallColormap (dpy, win->colormap);
+
+  /* Finally, give the window focus */
+  XSetInputFocus (dpy, win->w, 
+		  RevertToPointerRoot, CurrentTime);
+
+  XSync (dpy, False);
+}
+
+#if 0
+void
+unhide_transient_for (rp_window *win)
+{
+  rp_window_frame *frame;
+  rp_window *transient_for;
+
+  if (win == NULL) return;
+  if (!win->transient) return;
+  frame = find_windows_frame (win);
+
+  transient_for = find_window (win->transient_for);
+  if (transient_for == NULL) 
+    {
+      PRINT_DEBUG ("Can't find transient_for for '%s'", win->name );
+      return;
+    }
+
+  if (find_windows_frame (transient_for) == NULL)
+    {
+      set_frames_window (frame, transient_for);
+      maximize (transient_for);
+
+      PRINT_DEBUG ("unhide transient window: %s\n", transient_for->name);
+      
+      unhide_window_below (transient_for);
+
+      if (transient_for->transient) 
+	{
+	  unhide_transient_for (transient_for);
+	}
+
+      set_frames_window (frame, win);
+    }
+  else if (transient_for->transient) 
+    {
+      unhide_transient_for (transient_for);
+    }
+}
+#endif
+
+#if 0
+/* Hide all transient windows for win until we get to last. */
+void
+hide_transient_for_between (rp_window *win, rp_window *last)
+{
+  rp_window *transient_for;
+
+  if (win == NULL) return;
+  if (!win->transient) return;
+
+  transient_for = find_window (win->transient_for);
+  if (transient_for == last) 
+    {
+      PRINT_DEBUG ("Can't find transient_for for '%s'", win->name );
+      return;
+    }
+
+  if (find_windows_frame (transient_for) == NULL)
+    {
+      PRINT_DEBUG ("hide transient window: %s\n", transient_for->name);
+      hide_window (transient_for);
+    }
+
+  if (transient_for->transient) 
+    {
+      hide_transient_for (transient_for);
+    }
+}
+#endif
+
+#if 0
+void
+hide_transient_for (rp_window *win)
+{
+  /* Hide ALL the transient windows for win. */
+  hide_transient_for_between (win, NULL);
+}
+#endif
+
+#if 0
+/* return 1 if transient_for is a transient window for win or one of
+   win's transient_for ancestors. */
+int
+is_transient_ancestor (rp_window *win, rp_window *transient_for)
+{
+  rp_window *tmp;
+
+  if (win == NULL) return 0;
+  if (!win->transient) return 0;
+  if (transient_for == NULL) return 0;
+
+  tmp = win;
+
+  do
+    {
+      tmp = find_window (tmp->transient_for);
+      if (tmp == transient_for)
+	return 1;
+    } while (tmp && tmp->transient);
+  
+  return 0;
+}
+#endif
+
+void
+set_active_window (rp_window *win)
+{
+  rp_window *last_win;
+
+  if (win == NULL) return;
+
+  last_win = set_frames_window (rp_current_frame, win);
+
+  if (last_win) PRINT_DEBUG ("last window: %s\n", last_win->name);
+  PRINT_DEBUG ("new window: %s\n", win->name);
+
+  /* Make sure the window comes up full screen */
+  maximize (win);
+  unhide_window (win);
+
+#ifdef MAXSIZE_WINDOWS_ARE_TRANSIENTS
+  if (!win->transient
+      && !(win->hints->flags & PMaxSize 
+	   && win->hints->max_width < win->scr->root_attr.width
+	   && win->hints->max_height < win->scr->root_attr.height))
+#else
+  if (!win->transient)
+#endif
+    hide_others(win);
+
+  give_window_focus (win, last_win);
+
+  /* Make sure the program bar is always on the top */
+  update_window_names (win->scr);
+
+  XSync (dpy, False);
+}
+
+/* Go to the window, switching frames if the window is already in a
+   frame. */
+void
+goto_window (rp_window *win)
+{
+  rp_window_frame *frame;
+
+  frame = find_windows_frame (win);
+  if (frame)
+    {
+      set_active_frame (frame);
+    }
+  else
+    {
+      set_active_window (win);
+    }
 }
 
 void
-set_active_window (rp_window *rp_w)
+print_window_information (rp_window *win)
 {
-  static int counter = 1;	/* increments every time this function
-                                   is called. This way we can track
-                                   which window was last accessed.  */
+  marked_message_printf (0, 0, MESSAGE_WINDOW_INFORMATION, win->number, win->name);
+}
+
+/* get the window list and store it in buffer delimiting each window
+   with delim. mark_start and mark_end will be filled with the text
+   positions for the start and end of the current window. */
+void
+get_window_list (char *delim, struct sbuf *buffer, int *mark_start, int *mark_end)
+{
+  rp_window *w;
+  rp_window *other_window;
+  char dbuf[10];
+
+  if (buffer == NULL) return;
+
+  sbuf_clear (buffer);
+  other_window = find_window_other ();
+
+  for (w = rp_mapped_window_sentinel->next; 
+       w != rp_mapped_window_sentinel; 
+       w = w->next)
+    {
+      PRINT_DEBUG ("%d-%s\n", w->number, w->name);
+
+      if (w == current_window())
+	*mark_start = strlen (sbuf_get (buffer));
+
+      /* A hack, pad the window with a space at the beginning and end
+         if there is no delimiter. */
+      if (!delim)
+	sbuf_concat (buffer, " ");
+
+      sprintf (dbuf, "%d", w->number);
+      sbuf_concat (buffer, dbuf);
+
+      if (w == current_window())
+	sbuf_concat (buffer, "*");
+      else if (w == other_window)
+	sbuf_concat (buffer, "+");
+      else
+	sbuf_concat (buffer, "-");
 
-  if (rp_w == NULL) return;
+      sbuf_concat (buffer, w->name);
 
-  counter++;
-  rp_w->last_access = counter;
+      /* A hack, pad the window with a space at the beginning and end
+         if there is no delimiter. */
+      if (!delim)
+	sbuf_concat (buffer, " ");
 
-  if (rp_w->scr->bar_is_raised) update_window_names (rp_w->scr);
+      /* Only put the delimiter between the windows, and not after the the last
+         window. */
+      if (delim && w->next != rp_mapped_window_sentinel)
+	sbuf_concat (buffer, delim);
 
-  if (rp_current_window != NULL)
-    {
-      save_mouse_position (rp_current_window);
+      if (w == current_window())
+	*mark_end = strlen (sbuf_get (buffer));
     }
-
-  XWarpPointer (dpy, None, rp_w->scr->root, 
-		0, 0, 0, 0, rp_w->mouse_x, rp_w->mouse_y);
-  XSync (dpy, False);
 
-  XSetInputFocus (dpy, rp_w->w, 
-		  RevertToPointerRoot, CurrentTime);
-  XRaiseWindow (dpy, rp_w->w);
-  
-  /* Swap colormaps */
-  if (rp_current_window != NULL)
+  if (!strcmp (sbuf_get (buffer), ""))
     {
-      XUninstallColormap (dpy, rp_current_window->colormap);
+      sbuf_copy (buffer, MESSAGE_NO_MANAGED_WINDOWS);
     }
-  XInstallColormap (dpy, rp_w->colormap);
-
-  rp_current_window = rp_w;
-
-      
-  /* Make sure the program bar is always on the top */
-  update_window_names (rp_w->scr);
 }
