@@ -24,6 +24,7 @@
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
 #include <X11/Xproto.h>
+#include <X11/cursorfont.h>
 
 #include <errno.h>
 #include <stdio.h>
@@ -37,6 +38,9 @@
 
 static void init_screen (screen_info *s, int screen_num);
 
+int alarm_signalled = 0;
+int kill_signalled = 0;
+int hup_signalled = 0;
 int rat_x;
 int rat_y;
 int rat_visible = 1;		/* rat is visible by default */
@@ -50,6 +54,9 @@ Atom wm_colormaps;
 
 Atom rp_restart;
 Atom rp_kill;
+Atom rp_command;
+Atom rp_command_request;
+Atom rp_command_result;
 
 screen_info *screens;
 int num_screens;
@@ -67,13 +74,22 @@ struct rp_key prefix_key;
 
 struct modifier_info rp_modifier_info;
 
+/* rudeness levels */
+int rp_honour_transient_raise = 1;
+int rp_honour_normal_raise = 1;
+int rp_honour_transient_map = 1;
+int rp_honour_normal_map = 1;
+
 /* Command line options */
-static struct option ratpoison_longopts[] = { {"help", no_argument, 0, 'h'},
-					      {"version", no_argument, 0, 'v'},
-					      {"restart", no_argument, 0, 'r'},
-					      {"kill", no_argument, 0, 'k'},
-					      {0, 0, 0, 0} };
-static char ratpoison_opts[] = "hvrk";
+static struct option ratpoison_longopts[] = 
+  { {"help", 	no_argument, 		0, 	'h'},
+    {"version", no_argument, 		0, 	'v'},
+    {"restart", no_argument, 		0, 	'r'},
+    {"kill", 	no_argument, 		0, 	'k'},
+    {"command", required_argument, 	0, 	'c'},
+    {0, 	0, 			0, 	0} };
+
+static char ratpoison_opts[] = "hvrkc:";
 
 void
 fatal (const char *msg)
@@ -104,36 +120,19 @@ xrealloc (void *ptr, size_t size)
 void
 sighandler (int signum)
 {
-  fprintf (stderr, "ratpoison: Agg! I've been SHOT!\n"); 
-  clean_up ();
-  exit (EXIT_FAILURE);
+  kill_signalled++;
 }
 
 void
 hup_handler (int signum)
 {
-  /* Doesn't function correctly. The event IS placed on the queue but
-     XSync() doesn't seem to sync it and until other events come in
-     the restart event won't be processed. */
-  PRINT_DEBUG ("Restarting with a fresh plate.\n"); 
-
-  send_restart ();
-  XSync(dpy, False); 
+  hup_signalled++;
 }
 
 void
 alrm_handler (int signum)
 {
-  int i;
-
-  PRINT_DEBUG ("alarm recieved.\n");
-
-  /* FIXME: should only hide 1 bar, but we hide them all. */
-  for (i=0; i<num_screens; i++)
-    {
-      hide_bar (&screens[i]);
-    }
-  XSync (dpy, False);
+  alarm_signalled++;
 }
 
 int
@@ -205,7 +204,8 @@ print_help ()
   printf ("-h, --help            Display this help screen\n");
   printf ("-v, --version         Display the version\n");
   printf ("-r, --restart         Restart ratpoison\n");
-  printf ("-k, --kill            Kill ratpoison\n\n");
+  printf ("-k, --kill            Kill ratpoison\n");
+  printf ("-c, --command         Send ratpoison a colon-command\n\n");
 
   printf ("Report bugs to ratpoison-devel@lists.sourceforge.net\n\n");
 
@@ -249,7 +249,14 @@ read_rc_file (FILE *file)
 
 	  /* do it */
 	  if (*line != '#')
-	    command (line);
+	    {
+	      char *result;
+	      result = command (0, line);
+
+	      /* Gobble the result. */
+	      if (result)
+		free (result);
+	    }
 
 	  *line = '\0';
 	}
@@ -271,7 +278,9 @@ read_startup_files ()
 
   homedir = getenv ("HOME");
   if (!homedir)
-    fprintf (stderr, "ratpoison: $HOME not set!?\n");
+    {
+      PRINT_ERROR ("ratpoison: $HOME not set!?\n");
+    }
   else
     {
       char *filename = (char*)xmalloc (strlen (homedir) + strlen ("/.ratpoisonrc") + 1);
@@ -299,6 +308,43 @@ read_startup_files ()
     }
 }
 
+/* Odd that we spend so much code on making sure the silly welcome
+   message is correct. Oh well... */
+static void
+show_welcome_message ()
+{
+  rp_action *help_action;
+  char *prefix, *help;
+  
+  prefix = keysym_to_string (prefix_key.sym, prefix_key.state);
+
+  /* Find the help key binding. */
+  help_action = find_keybinding_by_action ("help");
+  if (help_action)
+    help = keysym_to_string (help_action->key, help_action->state);
+  else
+    help = NULL;
+
+
+  if (help)
+    {
+      /* A little kludge to use ? instead of `question' for the help
+	 key. */
+      if (!strcmp (help, "question"))
+	marked_message_printf (0, 0, MESSAGE_WELCOME, prefix, "?");
+      else
+	marked_message_printf (0, 0, MESSAGE_WELCOME, prefix, help);
+
+      free (help);
+    }
+  else
+    {
+      marked_message_printf (0, 0, MESSAGE_WELCOME, prefix, ":help");
+    }
+  
+  free (prefix);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -306,6 +352,8 @@ main (int argc, char *argv[])
   int c;
   int do_kill = 0;
   int do_restart = 0;
+  int do_command = 0;
+  char *command = NULL;
 
   myargv = argv;
 
@@ -331,6 +379,11 @@ main (int argc, char *argv[])
 	case 'r':
 	  do_restart = 1;
 	  break;
+	case 'c':
+	  command = xmalloc (strlen (optarg) + 1);
+	  strcpy (command, optarg);
+	  do_command = 1;
+	  break;
 	default:
 	  exit (EXIT_FAILURE);
 	}
@@ -342,29 +395,32 @@ main (int argc, char *argv[])
       return EXIT_FAILURE;
     }
 
-  /* Setup signal handlers. */
-  XSetErrorHandler(handler);  
-  set_sig_handler (SIGALRM, alrm_handler);
-  set_sig_handler (SIGTERM, sighandler);
-  set_sig_handler (SIGINT, sighandler);
-  set_sig_handler (SIGHUP, hup_handler);
-
   /* Set ratpoison specific Atoms. */
   rp_restart = XInternAtom (dpy, "RP_RESTART", False);
   rp_kill = XInternAtom (dpy, "RP_KILL", False);
+  rp_command = XInternAtom (dpy, "RP_COMMAND", False);
+  rp_command_request = XInternAtom (dpy, "RP_COMMAND_REQUEST", False);
+  rp_command_result = XInternAtom (dpy, "RP_COMMAND_RESULT", False);
 
   if (do_kill)
     {
       send_kill ();
       XSync (dpy, False);
-      clean_up ();
+      XCloseDisplay (dpy);
       return EXIT_SUCCESS;
     }
   if (do_restart)
     {
       send_restart ();
       XSync (dpy, False);
-      clean_up ();
+      XCloseDisplay (dpy);
+      return EXIT_SUCCESS;
+    }
+  if (do_command)
+    {
+      send_command (command);
+      free (command);
+      XCloseDisplay (dpy);
       return EXIT_SUCCESS;
     }
 
@@ -376,10 +432,19 @@ main (int argc, char *argv[])
   wm_take_focus = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
   wm_colormaps = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
 
+  /* Setup signal handlers. */
+  XSetErrorHandler(handler);  
+  set_sig_handler (SIGALRM, alrm_handler);
+  set_sig_handler (SIGTERM, sighandler);
+  set_sig_handler (SIGINT, sighandler);
+  set_sig_handler (SIGHUP, hup_handler);
+
+  /* Setup ratpoison's internal structures */
   init_numbers ();
   init_window_list ();
+  init_frame_list ();
+  update_modifier_map ();
   initialize_default_keybindings ();
-  init_modifier_map ();
 
   font = XLoadQueryFont (dpy, FONT);
   if (font == NULL)
@@ -401,24 +466,41 @@ main (int argc, char *argv[])
 
   read_startup_files ();
 
-  /* Set an initial window as active. */
-  set_active_window (find_window_other ());
+  /* Indicate to the user that ratpoison has booted. */
+  show_welcome_message();
   
   handle_events ();
 
   return EXIT_SUCCESS;
 }
 
+static void
+init_rat_cursor (screen_info *s)
+{
+  s->rat = XCreateFontCursor( dpy, XC_icon );
+}
+
 static void
 init_screen (screen_info *s, int screen_num)
 {
   XColor fg_color, bg_color,/*  bold_color, */ junk;
 
+  /* Select on some events on the root window, if this fails, then
+     there is already a WM running and the X Error handler will catch
+     it, terminating ratpoison. */
+  XSelectInput(dpy, RootWindow (dpy, screen_num),
+               PropertyChangeMask | ColormapChangeMask
+               | SubstructureRedirectMask | KeyPressMask | KeyReleaseMask
+               | SubstructureNotifyMask );
+  XSync (dpy, False);
+
   s->screen_num = screen_num;
   s->root = RootWindow (dpy, screen_num);
   s->def_cmap = DefaultColormap (dpy, screen_num);
   s->font = font;
   XGetWindowAttributes (dpy, s->root, &s->root_attr);
+  
+  init_rat_cursor (s);
 
   /* Get our program bar colors */
   if (!XAllocNamedColor (dpy, s->def_cmap, FOREGROUND, &fg_color, &junk))
@@ -431,11 +513,6 @@ init_screen (screen_info *s, int screen_num)
       fprintf (stderr, "ratpoison: Unknown color '%s'\n", BACKGROUND);
     }
 
-/*   if (!XAllocNamedColor (dpy, s->def_cmap, BAR_BOLD_COLOR, &bold_color, &junk)) */
-/*     { */
-/*       fprintf (stderr, "ratpoison: Unknown color '%s'\n", BAR_BOLD_COLOR); */
-/*     } */
-
   /* Setup the GC for drawing the font. */
   gv.foreground = fg_color.pixel;
   gv.background = bg_color.pixel;
@@ -447,16 +524,6 @@ init_screen (screen_info *s, int screen_num)
 			   GCForeground | GCBackground | GCFunction 
 			   | GCLineWidth | GCSubwindowMode | GCFont, 
 			   &gv);
-/*   gv.foreground = bold_color.pixel; */
-/*   s->bold_gc = XCreateGC(dpy, s->root,  */
-/* 			 GCForeground | GCBackground | GCFunction  */
-/* 			 | GCLineWidth | GCSubwindowMode | GCFont,  */
-/* 			 &gv); */
-
-  XSelectInput(dpy, s->root,
-               PropertyChangeMask | ColormapChangeMask
-               | SubstructureRedirectMask | KeyPressMask | KeyReleaseMask
-               | SubstructureNotifyMask );
 
   /* Create the program bar window. */
   s->bar_is_raised = 0;
@@ -475,6 +542,16 @@ init_screen (screen_info *s, int screen_num)
   					 1, 1, 1, fg_color.pixel, bg_color.pixel);
   XSelectInput (dpy, s->input_window, KeyPressMask );
 
+  /* Create the frame indicator window */
+  s->frame_window = XCreateSimpleWindow (dpy, s->root, 1, 1, 1, 1, 1, 
+					 fg_color.pixel, bg_color.pixel);
+  XSelectInput (dpy, s->frame_window, KeyPressMask );
+
+  /* Create the help window */
+  s->help_window = XCreateSimpleWindow (dpy, s->root, 0, 0, s->root_attr.width,
+					s->root_attr.height, 1, fg_color.pixel, bg_color.pixel);
+  XSelectInput (dpy, s->help_window, KeyPressMask);
+
   XSync (dpy, 0);
 
   scanwins (s);
@@ -483,6 +560,37 @@ init_screen (screen_info *s, int screen_num)
 void
 clean_up ()
 {
+  rp_window *cur;
+  int i;
+
+  /* Map any iconized windows. */
+  for (cur = rp_mapped_window_sentinel->next;
+       cur != rp_mapped_window_sentinel;
+       cur = cur->next)
+    {
+      if (cur->state == IconicState)
+	{
+	  unhide_window (cur);
+	}
+    }
+  
+  XSync (dpy, False);
+
+  for (i=0; i<num_screens; i++)
+    {
+      XDestroyWindow (dpy, screens[i].bar_window);
+      XDestroyWindow (dpy, screens[i].key_window);
+      XDestroyWindow (dpy, screens[i].input_window);
+      XDestroyWindow (dpy, screens[i].frame_window);
+      XDestroyWindow (dpy, screens[i].help_window);
+
+      XFreeCursor (dpy, screens[i].rat);
+      XFreeColormap (dpy, screens[i].def_cmap);
+      XFreeGC (dpy, screens[i].normal_gc);
+    }
+
+  XFreeFont (dpy, font);
+
   XSetInputFocus (dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
   XCloseDisplay (dpy);
 }
