@@ -48,8 +48,7 @@ grab_prefix_key (Window w)
   XGrabKey(dpy, AnyKey, AnyModifier, w, True, 
 	   GrabModeAsync, GrabModeAsync);
 #else
-  XGrabKey(dpy, XKeysymToKeycode (dpy, prefix_key.sym ), prefix_key.state, 
-	   w, True, GrabModeAsync, GrabModeAsync);
+  grab_key (XKeysymToKeycode (dpy, prefix_key.sym), prefix_key.state, w);
 #endif
 }
 
@@ -58,15 +57,15 @@ ungrab_prefix_key (Window w)
 {
 #ifdef HIDE_MOUSE
 #else
-  XUngrabKey(dpy, XKeysymToKeycode (dpy, prefix_key.sym ), prefix_key.state, w);
+  XUngrabKey(dpy, XKeysymToKeycode (dpy, prefix_key.sym), AnyModifier, w);
 #endif
 }
 
 screen_info*
 current_screen ()
 {
-  if (rp_current_window)
-    return rp_current_window->scr;
+  if (current_window())
+    return current_window()->scr;
   else
     return &screens[0];
 }
@@ -175,8 +174,6 @@ update_window_information (rp_window *win)
 {
   XWindowAttributes attr;
 
-  PRINT_DEBUG ("managing new window\n");
-
   update_window_name (win);
 
   /* Get the WM Hints */
@@ -189,6 +186,10 @@ update_window_information (rp_window *win)
   win->y = attr.y;
   win->width = attr.width;
   win->height = attr.height;
+  win->border = attr.border_width;
+
+  /* Transient status */
+  win->transient = XGetTransientForHint (dpy, win->w, &win->transient_for);  
 }
 
 void
@@ -196,16 +197,15 @@ unmanage (rp_window *w)
 {
   return_window_number (w->number);
   remove_from_list (w);
-  XFree (w->hints);
 
-  free (w);
+  free_window (w);
 
 #ifdef AUTO_CLOSE
   if (rp_mapped_window_sentinel->next == rp_mapped_window_sentinel
       && rp_mapped_window_sentinel->prev == rp_mapped_window_sentinel)
     {
       /* If the mapped window list is empty then we have run out of
- 	 managed windows, So kill ratpoison. */
+ 	 managed windows, so kill ratpoison. */
 
       /* FIXME: The unmapped window list may also have to be checked
 	 in the case that the only mapped window in unmapped and
@@ -232,21 +232,25 @@ scanwins(screen_info *s)
   for (i = 0; i < nwins; i++) 
     {
       XGetWindowAttributes(dpy, wins[i], &attr);
-      if (wins[i] == s->bar_window || wins[i] == s->key_window || wins[i] == s->input_window) continue;
+      if (wins[i] == s->bar_window 
+	  || wins[i] == s->key_window 
+	  || wins[i] == s->input_window
+	  || wins[i] == s->frame_window
+	  || wins[i] == s->help_window
+	  || attr.override_redirect == True
+	  || unmanaged_window (wins[i])) continue;
 
-      if (attr.override_redirect != True && !unmanaged_window (wins[i]))
+      win = add_to_window_list (s, wins[i]);
+
+      PRINT_DEBUG ("map_state: %d\n", attr.map_state);
+      if (attr.map_state == IsViewable) 
 	{
-	  win = add_to_window_list (s, wins[i]);
-
-	  PRINT_DEBUG ("map_state: %d\n", attr.map_state);
-	  if (attr.map_state == IsViewable) 
-	    {
-	      map_window (win);
-	    }
-	      
+	  win->state = NormalState;
+	  map_window (win);
 	}
     }
-  XFree((void *) wins);	/* cast is to shut stoopid compiler up */
+
+  XFree(wins);
 }
 
 int
@@ -270,17 +274,15 @@ unmanaged_window (Window w)
   return 0;
 }
 
-/* Set the state of the window.
-
-   FIXME: This is sort of broken. We should really record the state in
-   win->state and mimic the X states NormalState, WithdrawnState,
-   IconicState */
+/* Set the state of the window. */
 void
 set_state (rp_window *win, int state)
 {
   long data[2];
   
-  data[0] = (long)state;
+  win->state = state;
+
+  data[0] = (long)win->state;
   data[1] = (long)None;
 
   XChangeProperty (dpy, win->w, wm_state, wm_state, 32,
@@ -292,8 +294,14 @@ set_state (rp_window *win, int state)
 static void
 maximize_transient (rp_window *win)
 {
+  rp_window_frame *frame;
   int maxx, maxy;
 
+  /* Set the window's border */
+  win->border = WINDOW_BORDER_WIDTH;
+
+  frame = win->frame;
+
   /* Honour the window's maximum size */
   if (win->hints->flags & PMaxSize)
     {
@@ -304,31 +312,46 @@ maximize_transient (rp_window *win)
     {
       maxx = win->width;
       maxy = win->height;
+    }
 
-      /* Make sure we maximize to the nearest Resize Increment specified
-	 by the window */
-      if (win->hints->flags & PResizeInc)
-	{
-	  int amount;
+  /* Make sure we maximize to the nearest Resize Increment specified
+     by the window */
+  if (win->hints->flags & PResizeInc)
+    {
+      int amount;
+      int delta;
 
-	  amount = maxx - win->width;
-	  amount -= amount % win->hints->width_inc;
-	  if (amount < 0) amount -= win->hints->width_inc;
-	  PRINT_DEBUG ("amount x: %d\n", amount);
-	  maxx = amount + win->width;
+      amount = maxx - win->width;
+      delta = amount % win->hints->width_inc;
+      amount -= delta;
+      if (amount < 0 && delta) amount -= win->hints->width_inc;
+      maxx = amount + win->width;
 
-	  amount = maxy - win->height;
-	  amount -= amount % win->hints->height_inc;
-	  if (amount < 0) amount -= win->hints->height_inc;
-	  PRINT_DEBUG ("amount y: %d\n", amount);
-	  maxy = amount + win->height;
-	}
+      amount = maxy - win->height;
+      delta = amount % win->hints->height_inc;
+      amount -= delta;
+      if (amount < 0 && delta) amount -= win->hints->height_inc;
+      maxy = amount + win->height;
     }
 
   PRINT_DEBUG ("maxsize: %d %d\n", maxx, maxy);
 
-  win->x = PADDING_LEFT - win->width / 2 + (win->scr->root_attr.width - PADDING_LEFT - PADDING_RIGHT) / 2;;
-  win->y = PADDING_TOP - win->height / 2 + (win->scr->root_attr.height - PADDING_TOP - PADDING_BOTTOM) / 2;;
+  /* Fit the window inside its frame (if it has one) */
+  if (frame)
+    {
+      win->x = frame->x - win->width / 2 
+	+ (frame->width - win->border * 2) / 2;
+      win->y = frame->y - win->height / 2
+	+ (frame->height - win->border * 2) / 2;
+    }
+  else
+    {
+      win->x = PADDING_LEFT - win->width / 2 
+	+ (win->scr->root_attr.width - PADDING_LEFT - PADDING_RIGHT - win->border * 2) / 2;
+      win->y = PADDING_TOP - win->height / 2 
+	+ (win->scr->root_attr.height - PADDING_TOP - PADDING_BOTTOM - win->border * 2) / 2;
+    }
+
   win->width = maxx;
   win->height = maxy;
 }
@@ -338,50 +361,82 @@ maximize_transient (rp_window *win)
 static void
 maximize_normal (rp_window *win)
 {
+  rp_window_frame *frame;
   int maxx, maxy;
-
   int off_x = 0;
   int off_y = 0;
 
+  /* Set the window's border */
+  win->border = WINDOW_BORDER_WIDTH;
+
+  frame = win->frame;
+
   /* Honour the window's maximum size */
   if (win->hints->flags & PMaxSize)
     {
       maxx = win->hints->max_width;
       maxy = win->hints->max_height;
-      
-      /* centre the non-maximized window */
-/*       off_x = ((win->scr->root_attr.width - PADDING_LEFT - PADDING_RIGHT) - win->hints->max_width) / 2; */
-/*       off_y = ((win->scr->root_attr.height - PADDING_TOP - PADDING_BOTTOM) - win->hints->max_height) / 2; */
     }
   else
     {
-      maxx = win->scr->root_attr.width - PADDING_LEFT - PADDING_RIGHT;
-      maxy = win->scr->root_attr.height - PADDING_TOP - PADDING_BOTTOM;
+      maxx = win->scr->root_attr.width 
+	- PADDING_LEFT - PADDING_RIGHT - win->border * 2;
+      maxy = win->scr->root_attr.height 
+	- PADDING_TOP - PADDING_BOTTOM - win->border * 2;
+    }
 
-      /* Make sure we maximize to the nearest Resize Increment specified
-	 by the window */
-      if (win->hints->flags & PResizeInc)
-	{
-	  int amount;
+  /* Fit the window inside its frame (if it has one) */
+  if (frame)
+    {
+      PRINT_DEBUG ("frame width=%d height=%d\n", 
+		   frame->width, frame->height);
 
-	  amount = maxx - win->width;
-	  amount -= amount % win->hints->width_inc;
-	  if (amount < 0) amount -= win->hints->width_inc;
-	  PRINT_DEBUG ("amount x: %d\n", amount);
-	  maxx = amount + win->width;
+      if (maxx > frame->width) maxx = frame->width - win->border * 2;
+      if (maxy > frame->height) maxy = frame->height - win->border * 2;
+    }
 
-	  amount = maxy - win->height;
-	  amount -= amount % win->hints->height_inc;
-	  if (amount < 0) amount -= win->hints->height_inc;
-	  PRINT_DEBUG ("amount y: %d\n", amount);
-	  maxy = amount + win->height;
-	}
+  /* Make sure we maximize to the nearest Resize Increment specified
+     by the window */
+  if (win->hints->flags & PResizeInc)
+    {
+      int amount;
+      int delta;
+
+      amount = maxx - win->width;
+      delta = amount % win->hints->width_inc;
+      if (amount < 0 && delta) amount -= win->hints->width_inc;
+      amount -= delta;
+      maxx = amount + win->width;
+
+      amount = maxy - win->height;
+      delta = amount % win->hints->height_inc;
+      if (amount < 0 && delta) amount -= win->hints->height_inc;
+      amount -= delta;
+      maxy = amount + win->height;
     }
 
   PRINT_DEBUG ("maxsize: %d %d\n", maxx, maxy);
+ 
+#ifdef MAXSIZE_WINDOWS_ARE_TRANSIENTS
+  if (win->hints->flags & PMaxSize && frame)
+    {
+      off_x = (frame->width - win->border * 2) / 2 - maxx / 2;
+      off_y = (frame->height - win->border * 2) / 2 - maxy / 2;
+    }
+#endif
+
+  /* Fit the window inside its frame (if it has one) */
+  if (frame)
+    {
+      win->x = frame->x + off_x;
+      win->y = frame->y + off_y;
+    }
+  else
+    {
+      win->x = PADDING_LEFT + off_x;
+      win->y = PADDING_TOP + off_y;
+    }
 
-  win->x = PADDING_LEFT + off_x;
-  win->y = PADDING_TOP + off_y;
   win->width = maxx;
   win->height = maxy;
 }
@@ -391,11 +446,11 @@ maximize_normal (rp_window *win)
 void
 maximize (rp_window *win)
 {
-  if (!win) win = rp_current_window;
+  if (!win) win = current_window();
   if (!win) return;
 
-  /* Handle maximizing transient windows differently */
-  if (win->transient) 
+  /* Handle maximizing transient windows differently. */
+  if (win->transient)
     {
       maximize_transient (win);
     }
@@ -406,7 +461,7 @@ maximize (rp_window *win)
 
   /* Actually do the maximizing */
   XMoveResizeWindow (dpy, win->w, win->x, win->y, win->width, win->height);
-
+  XSetWindowBorderWidth (dpy, win->w, win->border);
   XSync (dpy, False);
 }
 
@@ -415,7 +470,7 @@ maximize (rp_window *win)
 void
 force_maximize (rp_window *win)
 {
-  if (!win) win = rp_current_window;
+  if (!win) win = current_window();
   if (!win) return;
 
   maximize_normal(win);
@@ -423,7 +478,7 @@ force_maximize (rp_window *win)
   /* Actually do the maximizing */
   XMoveResizeWindow (dpy, win->w, win->x, win->y, win->width+1, win->height+1);
   XMoveResizeWindow (dpy, win->w, win->x, win->y, win->width, win->height);
-
+  XSetWindowBorderWidth (dpy, win->w, win->border);
   XSync (dpy, False);
 }
 
@@ -433,19 +488,127 @@ map_window (rp_window *win)
 {
   PRINT_DEBUG ("Mapping the unmapped window %s\n", win->name);
 
+  /* Fill in the necessary data about the window */
   update_window_information (win);
-  grab_prefix_key (win->w);
-
-  maximize (win);
-
-  XMapWindow (dpy, win->w);
-  set_state (win, NormalState);
-  set_active_window (win);
-
-  win->state = STATE_MAPPED;
   win->number = get_unique_window_number ();
+  grab_prefix_key (win->w);
 
   /* Put win in the mapped window list */
   remove_from_list (win);
   insert_into_list (win, rp_mapped_window_sentinel);
+
+  /* The window has never been accessed since it was brought back from
+     the Withdrawn state. */
+  win->last_access = 0;
+
+  /* It is now considered iconic and set_active_window can handle the rest. */
+  set_state (win, IconicState);
+
+  /* Depending on the rudeness level, actually map the window. */
+  if ((rp_honour_transient_map && win->transient)
+      || (rp_honour_normal_map && !win->transient))
+    set_active_window (win);
+  else
+    {
+      if (win->transient)
+	marked_message_printf (0, 0, MESSAGE_MAP_TRANSIENT, 
+			       win->number, win->name);
+      else
+	marked_message_printf (0, 0, MESSAGE_MAP_WINDOW,
+			       win->number, win->name);
+    }
+}
+
+void
+hide_window (rp_window *win)
+{
+  if (win == NULL) return;
+
+  /* An unmapped window is not inside a frame. */
+  win->frame = NULL;
+
+  /* Ignore the unmap_notify event. */
+  XSelectInput(dpy, win->w, WIN_EVENTS&~(StructureNotifyMask));
+  XUnmapWindow (dpy, win->w);
+  XSelectInput (dpy, win->w, WIN_EVENTS);
+  set_state (win, IconicState);
+}
+
+void
+unhide_window (rp_window *win)
+{
+  if (win == NULL) return;
+
+  /* Always raise the window. */
+  XRaiseWindow (dpy, win->w);
+
+  if (win->state != IconicState) return;
+
+  XMapWindow (dpy, win->w);
+  set_state (win, NormalState);
+}
+
+/* same as unhide_window except that it makes sure the window is mapped
+   on the bottom of the window stack. */
+void
+unhide_window_below (rp_window *win)
+{
+  if (win == NULL) return;
+
+  /* Always lower the window, but if its not iconic we don't need to
+     map it since it already is mapped. */
+  XLowerWindow (dpy, win->w);
+
+  if (win->state != IconicState) return;
+
+  XMapWindow (dpy, win->w);
+  set_state (win, NormalState);
+}
+
+void
+withdraw_window (rp_window *win)
+{
+  if (win == NULL) return;
+
+  PRINT_DEBUG ("withdraw_window on '%s'\n", win->name);
+
+  /* Give back the window number. the window will get another one,
+     if it is remapped. */
+  return_window_number (win->number);
+  win->number = -1;
+
+  remove_from_list (win);
+  append_to_list (win, rp_unmapped_window_sentinel);
+
+  ignore_badwindow++;
+
+  XRemoveFromSaveSet (dpy, win->w);
+  set_state (win, WithdrawnState);
+  XSync (dpy, False);
+
+  ignore_badwindow--;
+}
+
+/* Hide all other mapped windows except for win in win's frame. */
+void
+hide_others (rp_window *win)
+{
+  rp_window_frame *frame;
+  rp_window *cur;
+
+  if (win == NULL) return;
+  frame = find_windows_frame (win);
+  if (frame == NULL) return;
+
+  for (cur = rp_mapped_window_sentinel->next; 
+       cur != rp_mapped_window_sentinel; 
+       cur = cur->next)
+    {
+      if (find_windows_frame (cur) 
+	  || cur->state != NormalState 
+	  || cur->frame != frame)
+	continue;
+
+      hide_window (cur);
+    }
 }
