@@ -43,4 +43,10 @@ void force_maximize (rp_window *win);
 void grab_prefix_key (Window w);
 void ungrab_prefix_key (Window w);
 
+void hide_window (rp_window *win);
+void unhide_window (rp_window *win);
+void unhide_window_below (rp_window *win);
+void withdraw_window (rp_window *win);
+void hide_others (rp_window *win);
+
 #endif /* ! _RATPOISION_MANAGE_H */
